from collections import defaultdict, Counter, namedtuple

from Bio import SeqIO
from Bio.Alphabet import generic_dna
import numpy as np

VERSION = '0.1.1'

HXB2_LOS_ALAMOS_ID = 'B.FR.83.HXB2_LAI_IIIB_BRU.K03455'


def parse_breakpoints(bp_files):
    """
    Parses jpHMM-compatible breakpoint file(s) (such as recombination.txt) into a dictionary.
    The file should have the following format:

    =====================
    # optional-comment
    # ...
    # optional-comment

    >sequence_id_1 optional-text-to-be-ignored
    start_position_1_1	end_position_1_1	subtype_1_1
    ...
    start_position_1_n1	end_position_1_n1	subtype_1_n1

    >sequence_id_2 optional-text-to-be-ignored
    start_position_2_1	end_position_2_1	subtype_2_1/subtype_2_2
    ...
    =====================

    The start and end breakpoint positions are inclusive, and 1-based.
    The start-end intervals should cover all the sequence,
    therefore start_position_i_1 is always 1 and end_position_i_ni is the length of the sequence i.
    If subtype_i_k is the same ad subtype_i_l then this subtype appears several times in this sequence
    (including intervals k and l).
    If several subtypes are specified for an interval, e.g. subtype_i_k/subtype_i_l,
    it means that any of them could correspond to this interval.


    :param bp_files: path to jpHMM-compatible breakpoint file(s) (such as recombination.txt)
    :type bp_files: list(str) or str
    :return: mapping between sequences ids and a mapping of subtypes to breakpoint interval list:
        {id: {subtype: [(start, stop), ...], ...}, ..}
    :rtype: dict
    """
    id2subtype2interval = {}
    if isinstance(bp_files, str):
        bp_files = [bp_files]
    for bp_file in bp_files:
        with open(bp_file, 'r') as f:
            name = None
            for line in f:
                line = line.strip('\n').strip()
                if not line or line.startswith('#'):
                    continue
                if line.startswith('>'):
                    if name:
                        id2subtype2interval[name] = subtype2interval
                    name = line.split(' ')[0].strip()[1:]
                    subtype2interval = defaultdict(list)
                else:
                    start, stop, subtypes = line.split('\t')
                    for subtype in subtypes.split('/'):
                        subtype2interval[subtype].append((int(start), int(stop)))
            if name:
                id2subtype2interval[name] = subtype2interval
    return id2subtype2interval


def breakpoints2bitmasks(id2subtype2interval, generalise_subtypes=True):
    """
    Converts breakpoint dictionary to a bitmask dictionary,
    with each subtype represented by a boolean array where True correspond to the subtype interval positions.
    For example, the following breakpoints
    >CRF
    1	3	A
    4	5	B
    correspond to the following bitmasks:
    A 1 1 1 0 0
    B 0 0 0 1 1

    :param id2subtype2interval: mapping between sequences ids and a mapping of subtypes to breakpoint interval list:
        {id: {subtype: [(start, stop), ...], ...}, ..}
    :type id2subtype2interval: dict
    :param generalise_subtypes: if True the more specific subtypes (e.g. A1, A2)
        will be replaced by more generic ones (e.g. A).
    :type generalise_subtypes: bool
    :return: mapping between sequences ids and a mapping of subtypes to bitmasks (represented as boolean numpy arrays):
        {id: {subtype: bitmask, ...}, ..}
    :rtype: dict
    """

    id2st2bitmask = {}
    for name, subtype2intervals in id2subtype2interval.items():
        subtype2bitmask = defaultdict(lambda: np.zeros(max(max(stop for (start, stop) in intervals)
                                                           for intervals in subtype2intervals.values()), dtype=bool))
        for subtype, intervals in subtype2intervals.items():
            if generalise_subtypes and len(subtype) == 2 and subtype[1].isdigit() and subtype[0].isalpha():
                subtype = subtype[0]
            for (start, stop) in intervals:
                subtype2bitmask[subtype][start - 1: stop] = 1
        id2st2bitmask[name] = subtype2bitmask
    return id2st2bitmask


def expand_crfs(id2subtype2bitmask, crf2st2bitmask):
    """
    Replaces CRFs bitmasks with their corresponding primary subtype bitmasks.

    For example if a CRF corresponds to the following recombination:
    >CRF
    1	3	A
    4	5	B
    and a sequence has the following bitmask mapping:
    A   1 0 0 0 0
    C   0 0 0 0 1
    CRF 0 1 1 1 0
    the replacement will transform it to:
    A   1 1 1 0 0
    B   0 0 0 1 0
    C   0 0 0 0 1


    :param id2subtype2bitmask: mapping between sequences ids and a mapping of subtypes to bitmasks
        (represented as boolean numpy arrays): {id: {subtype: bitmask, ...}, ..}
    :type id2subtype2bitmask: dict
    :param crf2st2bitmask: mapping between CRF ids and a mapping of subtypes to bitmasks
        (represented as boolean numpy arrays): {id: {subtype: bitmask, ...}, ..}
    :return: crf2st2bitmask: dict
    """
    for name, st2bm in id2subtype2bitmask.items():
        crfs = [_ for _ in st2bm.keys() if _.startswith('CRF')]
        while crfs:
            for crf in crfs:
                bm = st2bm[crf]
                for st, bitmask in crf2st2bitmask[crf].items():
                    intersection = bitmask & bm
                    if np.any(intersection):
                        st2bm[st] |= intersection
                del st2bm[crf]
            crfs = [_ for _ in st2bm.keys() if _.startswith('CRF')]


def get_reference_coordinates(aln_file, reference_id=HXB2_LOS_ALAMOS_ID):
    """
    Finds the coordinates (0-based) of reference sequence in the aligned file,
    e.g. if the reference sequence is AACCGT and it is aligned as --AA-CC-G-T-,
    its positions are [2, 3, 5, 6, 8, 10].
    Also returns the length of the initial sequence (n) and of the alignment (n_gappy).

    :param aln_file: path to the alignment file in fasta format.
    :type aln_file: str
    :param reference_id: id of the reference sequence in the alignment
    :type reference_id: str
    :return: Coordinates object with the filed coordinates containing the coordinates,
        n containing the reference sequence length (without gaps), and n_gappy containing the alignment length.
    :rtype: collections.namedtuple
    """
    ref = None
    for rec in SeqIO.parse(aln_file, 'fasta', alphabet=generic_dna):
        if reference_id == rec.id:
            ref = rec
            break

    gappy_seq = str(ref.seq)
    n_gappy = len(gappy_seq)
    n = len(gappy_seq.replace('-', ''))
    position_shift = [i for (i, c) in enumerate(gappy_seq) if c != '-']
    Coordinates = namedtuple('Coordinates', ['coordinates', 'n', 'n_gappy'])
    return Coordinates(coordinates=position_shift, n=n, n_gappy=n_gappy)


def parse_aligned_coordinates(jphmm_msas):
    """
    Parses the jpHMM alignment file(s) (alignment_to_msa.txt), into a dictionary containing a mapping of sequence ids to
    the coordinates (0-based) of them aligned,
    e.g. if the reference seq1 sequence is AACCGT and the alignment_to_msa.txt contains:
    >seq1
    3 4 6 7 9 11
    it means that it should be aligned as --AA-CC-G-T... (where ... contains as many gaps as the alignment length requires),
    and its parsed (0-based) positions are [2, 3, 5, 6, 8, 10].

    :param jphmm_msas: path to jpHMM-produced alignment file(s) (such as alignment_to_msa.txt)
    :type jphmm_msas: list(str) or str
    :return: sequence id to coordinate list mapping
    :rtype: dict
    """
    id2pos = {}
    if isinstance(jphmm_msas, str):
        jphmm_msas = [jphmm_msas]
    for in_msa in jphmm_msas:
        with open(in_msa, 'r') as f:
            id = None
            for line in f.readlines():
                line = line.strip('\n').strip()
                if line.startswith('#') or not line:
                    continue
                if line.startswith('>'):
                    id = line[1:].strip()
                    continue
                id2pos[id] = [(int(_) - 1) for _ in line.split(' ')]
    return id2pos


def shift_bitmask(id2subtype2bitmask, id2coordinates, n_gappy, gap_fill=1):
    """
    Realignes the bitmasks in a mapping according to specified coordinates
    (e.g. adds gaps to the bitmasks filled according to gap_fill parameter, either with 1s or 0s).

    For example, a sequence with a bitmask
    A   1 1 1 0 0
    B   0 0 0 1 1
    and the following (0-based) coordinates in the alignment of length (n_gappy) 10:
    2 3 4 7 8
    will get the realigned bitmask of (gaps filled with 0s):
    A   0 0 1 1 1 0 0 0 0 0
    B   0 0 0 0 0 0 0 1 1 0


    :param id2subtype2bitmask: mapping between sequences ids and a mapping of subtypes to bitmasks
        (represented as boolean numpy arrays): {id: {subtype: bitmask, ...}, ..}
    :type id2subtype2bitmask: dict
    :param id2coordinates: sequence id to its coordinate list (positions in the alignment) mapping
    :type id2coordinates: dict
    :param n_gappy: length of the alignment
    :type n_gappy: int
    :param gap_fill: what to fill the gaps in the aligned bitmasks with: 0 or 1
    :type gap_fill: int (0 or 1)
    :return: mapping between sequences ids and insertion lengths
        (specified by the number of coordinates equal to -1 (5'-Insertion) and n_gappy (3'-Insertion)),
        plus modifies the input bitmask mapping
    :rtype: dict
    """
    id2insertion_length = Counter()
    for name, subtype2bitmask in id2subtype2bitmask.items():
        shifted_positions = id2coordinates[name]
        insertion_start = 0
        for _ in shifted_positions:
            if -1 == _:
                insertion_start += 1
            else:
                break
        insertion_end = 0
        for _ in reversed(shifted_positions):
            if n_gappy == _:
                insertion_end += 1
            else:
                break
        id2insertion_length[name] += insertion_end + insertion_start
        shifted_positions = shifted_positions[insertion_start: len(shifted_positions) - insertion_end]
        for st, bitmask in list(subtype2bitmask.items()):
            gappy_bitmask = np.ones(n_gappy, dtype=bool) if gap_fill else np.zeros(n_gappy, dtype=bool)
            gappy_bitmask[shifted_positions] = bitmask[insertion_start: len(bitmask) - insertion_end]
            if not np.any(gappy_bitmask):
                del subtype2bitmask[st]
            else:
                subtype2bitmask[st] = gappy_bitmask
    return id2insertion_length


def save_bitmask(id2st2bitmask, out_file):
    """
    Saves a bitmask mapping to a file, in the following format:

    ========================
    >>sequence_id_1
    subtype_1	pos_1_1 pos_1_2 ... pos_1_n
    ...
    subtype_m	pos_m_1 pos_m_2 ... pos_m_n

    >>sequence_id_2
    ...
    ========================
    where pos_i_j is 1 if subtype_i is detected at the position j and 0 otherwise. All-zero masks are omitted.

    :param id2subtype2bitmask: mapping between sequences ids and a mapping of subtypes to bitmasks
        (represented as boolean numpy arrays): {id: {subtype: bitmask, ...}, ..}
    :type id2subtype2bitmask: dict
    :param out_file: path to the file where the bitmask should be saved.
    :type out_file: str
    """
    with open(out_file, 'w+') as f:
        for name in sorted(id2st2bitmask.keys()):
            f.write('>>{}\n'.format(name))
            subtype2bitmask = id2st2bitmask[name]
            for st in sorted(subtype2bitmask.keys()):
                f.write('{}\t{}\n'.format(st, ' '.join(str(int(_)) for _ in subtype2bitmask[st])))


def parse_bitmask(bp_file):
    """
    Reads a bitmask mapping from a file to a dictionary. The file should be in the following format:

    ========================
    >>sequence_id_1
    subtype_1	pos_1_1 pos_1_2 ... pos_1_n
    ...
    subtype_m	pos_m_1 pos_m_2 ... pos_m_n

    >>sequence_id_2
    ...
    ========================
    where pos_i_j is 1 if subtype_i is detected at the position j and 0 otherwise. All-zero masks are omitted.

    :param bp_file: path to the file where the bitmask mapping is stored.
    :type bp_file: str
    :return: mapping between sequences ids and a mapping of subtypes to bitmasks
        (represented as boolean numpy arrays): {id: {subtype: bitmask, ...}, ..}
    :rtype: dict
    """
    id2st2bitmask = {}
    with open(bp_file, 'r') as f:
        name = None
        for line in f:
            line = line.strip('\n').strip()
            if not line:
                continue
            if line.startswith('>>'):
                if name:
                    id2st2bitmask[name] = st2bps
                name = line[2:]
                st2bps = {}
            else:
                st, mask = line.split('\t')
                st2bps[st] = np.array([bool(int(_)) for _ in mask.split(' ')])
        if name:
            id2st2bitmask[name] = st2bps
    return id2st2bitmask




import os
from collections import defaultdict

import numpy as np
import pandas as pd

from jphmm_tools import parse_breakpoints, breakpoints2bitmasks, shift_bitmask, expand_crfs, \
    parse_aligned_coordinates, \
    parse_bitmask, get_reference_coordinates, HXB2_LOS_ALAMOS_ID, VERSION

DATA_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')
HIV1_BREAKPOINTS = os.path.join(DATA_DIR, 'HIV1.breakpoints')

COMPATIBLE_SUBTYPE_COL = 'compatible_subtypes'
SUBTYPE_JPHMM_COL = 'subtype_jpHMM'


def get_subtypes(jphmm_msas, jphmm_recs, gappy_breakpoint_bitmask_file=None,
                 breakpoint_file=HIV1_BREAKPOINTS, aln_file=None,
                 reference_id=HXB2_LOS_ALAMOS_ID, slack=0, generalise_subtypes=True):
    """
    Converts the jpHMM output files to a dataframe containing two columns: one with the subtypes detected by jpHMM,
    and the other with the compatible subtypes (e.g. including different CRFs).

    :param jphmm_msas: path(s) to jpHMM-produced alignment file(s) (such as alignment_to_msa.txt)
    :type jphmm_msas: list(str) or str
    :param jphmm_recs: path(s) to jpHMM-produced breakpoint file(s) (such as recombination.txt)
    :type jphmm_recs: list(str) or str
    :param gappy_breakpoint_bitmask_file: (optional) path to the file containing bitmask for the CRF breakpoints,
        aligned. If not specified, breakpoint_file and aln_file must be specified.
    :type gappy_breakpoint_bitmask_file: str
    :param breakpoint_file: (optional) path to the file CRF breakpoint interval file.
        Must be specified together with the alignment file used by jpHMM.
        Otherwise, gappy_breakpoint_bitmask_file must be specified.
    :type breakpoint_file: str
    :param aln_file: (optional) path to the alignment file used by jpHMM.
        Must be specified together if breakpoint_file is used to specify CRF breakpoints.
    :param reference_id: (optional) need to be specified if aln_file+breakpoint_file are used to specify CRF breakpoints.
        The id of the sequences used as the reference for the breakpoint coordinates (HXB2 most probably) in the aln_file.
    :type reference_id: str
    :param slack: (optional, default is 0) number of nucleotides for which the wrong subtype
        can be ignored while matching CRFs.
    :type slack: int
    :param generalise_subtypes: if True the more specific subtypes (e.g. A1, A2)
        will be replaced by more generic ones (e.g. A).
    :type generalise_subtypes: bool
    :return: dataframe with sequence ids as indices and two columns,
        corresponding to the jpHMM detected subtypes inside each sequence and to the compatible subtypes and CRFs.
    :rtype: pandas.DataFrame
    """

    if gappy_breakpoint_bitmask_file is None:
        if breakpoint_file is None or aln_file is None:
            raise ValueError('Either the gappy breakpoint bitmask file '
                             'or the breakpoint file + the alignment file (used for jpHMM subtyping) must be specified')
        crf2st2bitmask, n_gappy = get_gappy_breakpoints(breakpoint_file, aln_file, reference_id,
                                                        generalise_subtypes=generalise_subtypes)
    else:
        crf2st2bitmask = parse_bitmask(gappy_breakpoint_bitmask_file)
        n_gappy = get_length(crf2st2bitmask)

    id2insertion_length, id2n, id2st2bitmask = jphmm2bitmask(crf2st2bitmask, jphmm_msas, jphmm_recs, n_gappy,
                                                             generalise_subtypes=generalise_subtypes)

    df = pd.DataFrame(columns=[SUBTYPE_JPHMM_COL, COMPATIBLE_SUBTYPE_COL])
    for name, st2bitmask in id2st2bitmask.items():
        length = id2n[name]
        ins_length = id2insertion_length[name]
        if length < ins_length:
            df.loc[name, SUBTYPE_JPHMM_COL] = None
            continue

        subtypes = set(st2bitmask.keys())
        df.loc[name, SUBTYPE_JPHMM_COL] = ','.join(sorted(subtypes))

        potential_subtypes = []
        if len(subtypes) == 1:
            potential_subtypes.extend(subtypes)

        for crf, crf_st2bitmask in crf2st2bitmask.items():
            if subtypes - set(crf_st2bitmask.keys()):
                continue
            incompatible = False
            for st, bitmask in st2bitmask.items():
                if np.any(bitmask & ~crf_st2bitmask[st]):
                    if slack:
                        incompatible_len = sum((bitmask & ~crf_st2bitmask[st]).astype(int))
                        if incompatible_len <= slack:
                            continue
                    incompatible = True
                    break
            if not incompatible:
                potential_subtypes.append(crf)
        df.loc[name, COMPATIBLE_SUBTYPE_COL] = '/'.join(sorted(potential_subtypes))

    return df


def get_gappy_breakpoints(breakpoint_file, aln_file, reference_id=HXB2_LOS_ALAMOS_ID, generalise_subtypes=True):
    """
    Converts a CRF breakpoint interval file and the alignment file used by jpHMM
    into an aligned breakpoint bitmask mapping.

    :param breakpoint_file: (optional) path to the file CRF breakpoint interval file.
        Must be specified together with the alignment file used by jpHMM.
        Otherwise, gappy_breakpoint_bitmask_file must be specified.
    :type breakpoint_file: str
    :param aln_file: (optional) path to the alignment file used by jpHMM.
        Must be specified together if breakpoint_file is used to specify CRF breakpoints.
    :param reference_id: (optional) need to be specified if aln_file+breakpoint_file are used to specify CRF breakpoints.
        The id of the sequences used as the reference for the breakpoint coordinates (HXB2 most probably) in the aln_file.
    :type reference_id: str
    :param generalise_subtypes: if True the more specific subtypes (e.g. A1, A2)
        will be replaced by more generic ones (e.g. A).
    :type generalise_subtypes: bool
    :return: mapping between CRF ids and a mapping of subtypes to breakpoint interval list (aligned):
        {id: {subtype: [(start, stop), ...], ...}, ..}
    :rtype: dict
    """
    crf2st2bitmask = breakpoints2bitmasks(parse_breakpoints(breakpoint_file), generalise_subtypes=generalise_subtypes)
    expand_crfs(crf2st2bitmask, crf2st2bitmask)

    position_shift, n, n_gappy = get_reference_coordinates(aln_file, reference_id=reference_id)
    shift_bitmask(crf2st2bitmask, defaultdict(lambda: position_shift), n_gappy)

    return crf2st2bitmask, n_gappy


def jphmm2bitmask(crf2st2bitmask, jphmm_msas, jphmm_recs, n_gappy, generalise_subtypes=True):
    """
    Converts a jpHMM output breakpoint interval (e.g. recombination.txt) and alignment (e.g. aignment_to_msa.txt) file(s)
    into an aligned breakpoint bitmask mapping.

    :param crf2st2bitmask: mapping between CRF ids and a mapping of subtypes to breakpoint interval list (aligned):
        {id: {subtype: [(start, stop), ...], ...}, ..}
    :type crf2st2bitmask: dict
    :param jphmm_msas: path to jpHMM-produced alignment file(s) (such as alignment_to_msa.txt)
    :type jphmm_msas: list(str) or str
    :param jphmm_recs: path to jpHMM-produced breakpoint file(s) (such as recombination.txt)
    :type jphmm_recs: list(str) or str
    :param n_gappy: length of the alignment
    :type n_gappy: int
    :param generalise_subtypes: if True the more specific subtypes (e.g. A1, A2)
        will be replaced by more generic ones (e.g. A).
    :type generalise_subtypes: bool
    :return: tuple of mappings of sequence ids to (1) lengths of (5'- plus 3'-) insertions detected by jpHMM,
        (2) lengths of the sequences, (3) mapping of subtypes to breakpoint interval list (aligned):
        {subtype: [(start, stop), ...], ...}
    :rtype: tuple(dict)
    """
    id2st2interval = parse_breakpoints(jphmm_recs)
    id2n = {name: max(max(stop for (start, stop) in intervals) for intervals in st2intervals.values())
            for name, st2intervals in id2st2interval.items()}
    id2st2bitmask = breakpoints2bitmasks(id2st2interval, generalise_subtypes=generalise_subtypes)
    id2pos = parse_aligned_coordinates(jphmm_msas)
    id2insertion_length = shift_bitmask(id2st2bitmask, id2pos, n_gappy, gap_fill=0)
    expand_crfs(id2st2bitmask, crf2st2bitmask)
    return id2insertion_length, id2n, id2st2bitmask


def get_length(crf2subtype2bitmask):
    """
    Calculates the alignment length for the breakpoint bitmask mapping.

    :param crf2subtype2bitmask: mapping between CRF ids and a mapping of subtypes to breakpoint interval list (aligned):
        {id: {subtype: [(start, stop), ...], ...}, ..}
    :type crf2subtype2bitmask: dict
    :return: alignment length
    :rtype: int
    """
    return len(next(iter(next(iter(crf2subtype2bitmask.values())).values())))


def main():
    """
    Entry point, calling :py:func:`jphmm_tools.aligner.get_subtypes` with command-line arguments,
    then saving the result to a file.

    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(
        description="Extracts sequence subtype information from the jpHMM (http://jphmm.gobics.de) output.",
        prog='jphmm_subtype')

    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=VERSION))

    jphmm_group = parser.add_argument_group('jpHMM-output-file-related arguments')
    jphmm_group.add_argument('--jphmm_msas', required=True, nargs='+', type=str,
                             help="Path(s) to jpHMM-produced alignment file(s) (such as alignment_to_msa.txt).")
    jphmm_group.add_argument('--jphmm_recs', required=True, nargs='+', type=str,
                             help="Path(s) to jpHMM-produced breakpoint file(s) (such as recombination.txt).")

    bp_group = parser.add_argument_group('CRF-breakpoint-related arguments')
    bp_group.add_argument('--gappy_breakpoint_bitmask_file', required=False, type=str,
                          help="Path to the file containing bitmask for the CRF breakpoints, aligned."
                               " If not specified, --breakpoint_file and --aln_file must be specified.")
    bp_group.add_argument('--breakpoint_file', required=False, default=HIV1_BREAKPOINTS, type=str,
                          help="Path to the file CRF breakpoint interval file."
                               " Must be specified together with the alignment file used by jpHMM."
                               " Otherwise, --gappy_breakpoint_bitmask_file must be specified.")
    bp_group.add_argument('--aln_file', required=False, type=str,
                          help="Path to the alignment file used by jpHMM."
                               " Must be specified together if --breakpoint_file is used to specify CRF breakpoints.")
    bp_group.add_argument('--reference_id', required=False, default=HXB2_LOS_ALAMOS_ID, type=str,
                          help="Need to be specified if --aln_file + --breakpoint_file are used to specify CRF breakpoints."
                               " The id of the sequences used as the reference for the breakpoint coordinates "
                               "(HXB2 most probably) in the --aln_file")

    tune_group = parser.add_argument_group('Fine-tuning arguments')
    tune_group.add_argument('--slack', required=False, default=0, type=int,
                            help="Number of nucleotides for which the wrong subtype can be ignored while matching CRFs.")
    tune_group.add_argument('--generalise_subtypes', action='store_true',
                            help="Replace the more specific subtypes (e.g. A1, A2) by more generic ones (e.g. A).")

    out_group = parser.add_argument_group('Output arguments')
    out_group.add_argument('--out_tab', required=True, type=str,
                           help="Path to the file where the tab-separated table "
                                "with the subtype infromation will be stored.")

    params = parser.parse_args()
    tab = params.out_tab

    params = vars(params)
    del params['out_tab']
    get_subtypes(**params).to_csv(tab, sep='\t', index_label='id')


if '__main__' == __name__:
    main()



import re
from collections import defaultdict

from Bio import SeqIO
from Bio.Alphabet import generic_dna

from jphmm_tools import VERSION


def get_subtype(name):
    """
    Extracts the subtype information from the sequence name,
    in subtype.whatever format (x.whatever for unknown subtypes.

    :param name: sequence name
    :type name: str
    :return: list of subtypes
    :rtype: list(str)
    """
    st = re.findall(r'^[^.]+[\._]', name)[0]
    st = st[:-1]
    if st == 'x':
        return []
    return ['CRF_{}'.format(_) if _[0].isdigit() else _ for _ in re.findall(r'[\d]{2}|A\d|F\d|[ABCDEFGHJKUO]', st)]


def reformat_reference(la_alignment, jphmm_ref_alignment):
    """
    Reformats a Los Alamos (https://www.hiv.lanl.gov/content/sequence/HIV/mainpage.html) alignment file
    with sequences name as subtype.whatever (x.whatever for unknown subtypes)
    into a reference file suitable for jpHMM (http://jphmm.gobics.de).

    :param la_alignment: path to the Los Alamos alignment file in fasta format.
    :type la_alignment: str
    :param jphmm_ref_alignment: path where the reference file suitable for jpHMM will be saved.
    :type jphmm_ref_alignment: str
    """
    subtype2rec = defaultdict(list)
    for rec in SeqIO.parse(la_alignment, 'fasta', alphabet=generic_dna):
        sts = get_subtype(rec.id)
        if len(sts) == 1:
            print(sts)
            subtype2rec[sts[0]].append(rec)
    with open(jphmm_ref_alignment, 'w') as f:
        for subtype in sorted(subtype2rec.keys()):
            f.write('>>{}\n'.format(subtype))
            for rec in subtype2rec[subtype]:
                f.write('>{}\n{}\n'.format(rec.id, rec.seq))


def main():
    """
    Entry point, calling :py:func:`jphmm_tools.la_referencer.reformat_reference` with command-line arguments.

    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(
        description="Reformats a Los Alamos (https://www.hiv.lanl.gov/content/sequence/HIV/mainpage.html) alignment file"
                    "into a reference file suitable for jpHMM (http://jphmm.gobics.de).",
        prog='jphmm_ref')

    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=VERSION))

    parser.add_argument('--la_alignment', required=True, type=str,
                        help='Path to the Los Alamos alignment file in fasta format, '
                             'with sequences name as subtype.whatever (x.whatever for unknown subtypes).')
    parser.add_argument('--jphmm_ref_alignment', required=True, type=str,
                        help='Path where the reference file suitable for jpHMM will be saved.')

    params = parser.parse_args()

    reformat_reference(**vars(params))


if '__main__' == __name__:
    main()



import numpy as np
from Bio import SeqIO
from Bio.Alphabet import generic_dna

from jphmm_tools.__init__ import parse_aligned_coordinates, VERSION


def align(jphmm_msas, in_fas, out_aln, aln_len):
    """
    Aligns input sequences according to the jpHMM (http://jphmm.gobics.de) output.

    :param jphmm_msas: path(s) to jpHMM-produced alignment file(s) (such as alignment_to_msa.txt)
    :type jphmm_msas: list(str) or str
    :param in_fas: path(s) to fasta file(s) containing sequences to be aligned
    :type in_fas: str
    :param out_aln: path where to save the output alignment file (in fasta format)
    :type out_aln: str
    :param aln_len: length of the alignment
    :type aln_len: int
    """

    id2pos = parse_aligned_coordinates(jphmm_msas)

    with open(out_aln, 'w') as f:
        if isinstance(in_fas, str):
            in_fas = [in_fas]
        for in_fa in in_fas:
            for rec in SeqIO.parse(in_fa, 'fasta', alphabet=generic_dna):
                positions = id2pos[rec.id]
                if positions:
                    aligned_seq = np.array(['-'] * aln_len)
                    seq = list(str(rec.seq))
                    insertion_beginning = sum(1 for _ in positions if _ == -1)
                    insertion_end = sum(1 for _ in positions if _ == aln_len)
                    aligned_seq[positions[insertion_beginning: (len(positions) - insertion_end)]] \
                        = seq[insertion_beginning: (len(seq) - insertion_end)]
                    f.write('>{}\n{}\n'.format(rec.id, ''.join(aligned_seq)))


def main():
    """
    Entry point, calling :py:func:`jphmm_tools.aligner.align` with command-line arguments.

    :return: void
    """
    import argparse

    parser = argparse.ArgumentParser(
        description="Aligns input sequences according to the jpHMM (http://jphmm.gobics.de) output.",
        prog='jphmm_align')

    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=VERSION))

    parser.add_argument('--jphmm_msas', required=True, nargs='+', type=str,
                        help="Path(s) to jpHMM-produced alignment file(s) (such as alignment_to_msa.txt).")
    parser.add_argument('--in_fas', required=True, nargs='+', type=str,
                        help="Path(s) to fasta file(s) containing sequences to be aligned.")
    parser.add_argument('--out_aln', required=True, type=str,
                        help="Path where to save the output alignment file (in fasta format).")
    parser.add_argument('--aln_len', required=True, type=int,
                        help="Length of the alignment.")

    params = parser.parse_args()

    align(**vars(params))


if '__main__' == __name__:
    main()



import os
from setuptools import setup, find_packages

setup(
    name='jphmm_tools',
    packages=find_packages(),
    include_package_data=True,
    package_data={'jphmm_tools': [os.path.join('data', '*.breakpoints'), os.path.join('..', 'README.md')]},
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    version='0.1.1',
    description='Tools for extracting information from jpHMM (http://jphmm.gobics.de) output.',
    author='Anna Zhukova',
    author_email='anna.zhukova@pasteur.fr',
    url='https://github.com/evolbioinfo/jphmm_tools',
    keywords=['jpHMM', 'HIV', 'subtyping', 'recombinants'],
    install_requires=['pandas', 'numpy', 'biopython'],
    entry_points={
            'console_scripts': [
                'jphmm_align = jphmm_tools.aligner:main',
                'jphmm_subtype = jphmm_tools.subtyper:main',
                'jphmm_ref = jphmm_tools.la_referencer:main'
            ]
    },
)
